#include <math.h>
#include <stdlib.h>
#include <memory.h>
#include <assert.h>
#include "skClustering.h"

#include "skOutput.h"

///////////////////
// skMeanShiftClustering1D()
//	Author: Cheng-en Guo, 01-26-2002
//	the standard mean shift clustering on 1-D
//	input:	pdData - pointer to the sample data, dim=nNum
//			nNum - data num
//			dWindowSize - the Parzen window size, which determine how big the cluster will covering
//	output:	pnClusterIndex - the cluster index for each sample, should have the dimension of nNum
//			pdClusterCenter - the mean value for each cluster, it won't be computed if not provide valid pointer
//	return: nClusterNum: succeeded0: failed, 
int skMeanShiftClustering1D(double *pdData, int nNum, double dWindowSize, int *pnClusterIndex, double *pdClusterCenter)
{
	if (nNum <= 0)
		return 0;

	int i, j;

	// initialize the cluster index
	for (i = 0; i < nNum; i++)
		pnClusterIndex[i] = i;

	// allocate the new memory for the data, because the data will be updated
	double *pdDataNew = new double[nNum];
	memcpy(pdDataNew, pdData, sizeof(double)*nNum);

	// update the data
	double aver, delta=0.00001;
	double sum;
	int n;
	while (true)
	{
		bool bUpdate = false;
		for (i = 0; i < nNum; i++)
		{
			n = 0;
			sum = 0;
			for (j = 0; j < nNum; j++)
			{
				if (fabs(pdDataNew[i] - pdDataNew[j]) <= dWindowSize/2)
				{
					sum += pdDataNew[j];
					n++;
				}
			}
			assert(n > 0);
			aver = sum / n;
			if (fabs(pdDataNew[i] - aver) > delta)
			{
				pdDataNew[i] = aver;
				bUpdate = true;
			}
		}
		if (!bUpdate)
			break;
	}

	int nClusterNum = 0;
	bool bFind;
	for (i = 0; i < nNum; i++)
	{
		bFind = false;
		for (j = 0; j < nClusterNum; j++)
		{
			if (fabs(pdDataNew[i] - pdClusterCenter[j]) <= 3*delta)
			{
				pnClusterIndex[i] = j;
				bFind = true;
				break;
			}
		}
		if (!bFind)
		{
			pnClusterIndex[i] = nClusterNum;
			pdClusterCenter[nClusterNum] = pdDataNew[i];
			nClusterNum++;
		}
	}

	delete [] pdDataNew;
	return nClusterNum;
}


///////////////////
// skKMeanCluster()
//	Author: Cheng-en Guo, 01-26-2002
//	input:	pdData - input data, nNum*nDim
//			nDim - sample dimension
//			nNum - sample num
//			pdWeight - weight for each dimension when calculate skEuclidianDistance
//			nClusterNum - supposed cluster num
//			nMaxClusterNum - supposed max cluster num
//	output:	pdCC - cluster center, should be nClusterNum*nDim dimension
//			nClusterNum - tree cluster num
//			pnCluster - final index of cluster for each sample, should be nNum*1 dimension
//	return: cluster num if successed, otherwise 0
int skKMeanCluster(double *pdData, int nDim, int nNum, double *pdWeight, double *pdCC, int &nClusterNum, int nMaxClusterNum, int *pnCluster)
{
	if (nDim <= 0 || nNum <= 0)
	{ 
		nClusterNum = 0;
		return 0;
	}

	if (nClusterNum <= 0)
		nClusterNum = nMaxClusterNum;

	if (nNum < 5)
		nClusterNum = 1;
	
	if (nNum < nClusterNum)
		nClusterNum = 1;

	int i;//index of sample
	int d;//index of dimension
	int n;//index of data
	int c;//index of cluster

	// if only one cluster desired
	if (nClusterNum ==  1)
	{
		for (d = 0; d < nDim; d++)
			pdCC[d] = 0;
		n  = 0;
		for (i = 0; i < nNum; i++)
			for (d = 0; d < nDim; d++)
				pdCC[d] += pdData[n++];
		for (d = 0; d < nDim; d++)
			pdCC[d] /= nNum;
		for (i = 0; i < nNum; i++)
			pnCluster[i] = 0;//only one cluster
		return 1;
	}

	// initilize the the cluster center
	// This is a greedy method. Different init may lead to different results
	for (c = 0; c < nClusterNum; c++)
	{
		i = (int)(rand()/(RAND_MAX+1.0) * nNum);
		for (d = 0; d < nDim; d++)
			pdCC[c*nDim + d] = pdData[i*nDim + d];
	}

//	memset(pnCluster, -1, sizeof(int)*nNum);
	// assign each sample to a cluster
	while (true)
	{
		if (skAssignSample2Cluster(pdData, nDim, nNum, pdWeight, pdCC, nClusterNum, pnCluster) <= 0)
			break;
		skUpdateClusterCenter(pdData, nDim, nNum, pdCC, nClusterNum, pnCluster);
	}
	return nClusterNum;
}


///////////////////
// skRobustCluster()
//	Author: Cheng-en Guo, 01-26-2002
//	Modified: CG, 05-16-2002
//	input:	pdData - input data, nNum*nDim
//			nDim - sample dimension
//			nNum - sample num
//			pdWeight - weight for each dimension when calculate skEuclidianDistance
//			nClusterNum - supposed cluster num
//			nMaxClusterNum - supposed max cluster num
//			nMinClusterSize - minimum the sample number in one cluster
//			dMinClusterDisRate - the DisRate=ClusterDis/MaxClusterDis, remove those cluster with low DisRate,
//								If set to 0.0, no consideration
//	output:	pdCC - cluster center, should be nClusterNum*nDim dimension
//			nClusterNum - tree cluster num
//			pnCluster - final index of cluster for each sample, should be nNum*1 dimension
//	return: cluster num if successed, otherwise 0
int skRobustCluster(double *pdData, int nDim, int nNum, double *pdWeight, double *pdCC, 
					int nMaxClusterNum, int nMinClusterSize, double dMinClusterDisRate, 
					int *pnCluster)
{
	int nClusterNum = nMaxClusterNum;

	if (nDim <= 0 || nNum <= 0)
	{ 
		return 0;
	}

	if (nClusterNum <= 0)
		nClusterNum = nMaxClusterNum;

	if (nNum < nMinClusterSize)
		nClusterNum = 1;
	
	if (nNum < nClusterNum)
		nClusterNum = 1;

	int i;//index of sample
	int d;//index of dimension
	int n;//index of data
	int c;//index of cluster

	// if only one cluster desired
	if (nClusterNum ==  1)
	{
		for (d = 0; d < nDim; d++)
			pdCC[d] = 0;
		n  = 0;
		for (i = 0; i < nNum; i++)
			for (d = 0; d < nDim; d++)
				pdCC[d] += pdData[n++];
		for (d = 0; d < nDim; d++)
			pdCC[d] /= nNum;
		for (i = 0; i < nNum; i++)
			pnCluster[i] = 0;//only one cluster
		return 1;
	}

/*
	// initilize the the cluster center
	// This is a greedy method. Different init may lead to different results
	for (c = 0; c < nClusterNum; c++)
	{
		i = (int)(rand()/(RAND_MAX+1.0) * nNum);
		for (d = 0; d < nDim; d++)
			pdCC[c*nDim + d] = pdData[i*nDim + d];
	}
*/
	int pnUsedSample[1024];
	c = 0;//assigned the cluster center
	i = (int)(rand()/(RAND_MAX+0.1) * nNum);
	pnUsedSample[c] = i;
	for (d = 0; d < nDim; d++)
		pdCC[c*nDim + d] = pdData[i*nDim + d];
	int max_no;
	double max_dis = 0, dis;
	for (c = 1; c < nClusterNum; c++)
	{
		max_dis = 0;
		max_no = -1;
		for (n = 0; n < nNum; n++)
		{
			dis = 0;
			for (int c0 = 0; c0 < c; c0++)
			{
				dis += skEuclidianDistance(pdData+n*nDim, pdCC+c0*nDim, nDim, pdWeight);
			}
			if (max_dis < dis)
			{
				bool bUsed = false;
				for (int c1 = 0; c1 < c; c1++)
				{
					if (pnUsedSample[c1] == n)
					{
						bUsed = true;
						break;
					}
				}
				if (!bUsed)
				{
					max_dis = dis;
					max_no = n;
				}
			}
		}
		assert(max_no >= 0);
		pnUsedSample[c] = max_no;
		for (d = 0; d < nDim; d++)
			pdCC[c*nDim + d] = pdData[max_no*nDim + d];
	}

	int	   *pnSampleNum = new int[nMaxClusterNum];//the sample num for each cluster
	memset(pnCluster, -1, sizeof(int)*nNum);
	// assign each sample to a cluster
	int *pnTest = new int[nClusterNum];
	//combine short skEuclidianDistance clusters
	double **ppdClusterDis = new double*[nMaxClusterNum];
	for (c = 0; c < nClusterNum; c++)
	{
		ppdClusterDis[c] = new double[nClusterNum];
		memset(ppdClusterDis[c], 0, sizeof(double)*nClusterNum);
	}
	while (true)
	{//with fixed cluster num update the samples until converge
		if (skAssignSample2Cluster(pdData, nDim, nNum, pdWeight, pdCC, nClusterNum, pnCluster) <= 0)
			break;
		skUpdateClusterCenter(pdData, nDim, nNum, pdCC, nClusterNum, pnCluster);
	}
	while (true)
	{
		// recompute the sample num in each cluster
		memset(pnSampleNum, 0, sizeof(int)*nMaxClusterNum);
		for (n = 0; n < nNum; n++)
		{
			pnSampleNum[pnCluster[n]]++;
		}
		// remove small size cluster
		bool bSmallCluster = false;
		for (c = 0; c < nClusterNum; c++)
		{
			if (pnSampleNum[c] < nMinClusterSize)
			{
				memcpy(pdCC+nDim*c, pdCC+nDim*(nClusterNum-1), sizeof(double)*nDim);//copy cluster center
				for (n = 0; n < nNum; n++)
				{
					if (pnCluster[n] == nClusterNum - 1)
						pnCluster[n] = c;
					if (pnCluster[n] == c)
						pnCluster[n] = -1;
				}
				c--;
				nClusterNum--;
				bSmallCluster = true;
				break;
			}
		}
		if (bSmallCluster)
		{
			while (true)
			{//with fixed cluster num update the samples until converge
				if (skAssignSample2Cluster(pdData, nDim, nNum, pdWeight, pdCC, nClusterNum, pnCluster) <= 0)
					break;
				skUpdateClusterCenter(pdData, nDim, nNum, pdCC, nClusterNum, pnCluster);
			}
		}
		// compute all the distances and find the min and max
		double dDis, dMaxDis = -10000000;
		double dMinDis = 1000000000000;
		int c1;
		for (c = 0; c < nClusterNum-1; c++)
		{
			for (c1 = c+1; c1 < nClusterNum; c1++)
			{
				dDis = skEuclidianDistance(pdCC+c*nDim, pdCC+c1*nDim, nDim, pdWeight);
				ppdClusterDis[c][c1] = dDis;
				if (dMaxDis < dDis)
					dMaxDis = dDis;
				if (dMinDis > dDis)
					dMinDis = dDis;
			}
		}

		double dDisTh = (dMaxDis)*dMinClusterDisRate;
		memset(pnTest, 0, sizeof(int)*nClusterNum);
		int nClusterNum_bak = nClusterNum;
		for (c = 0; c < nClusterNum-1; c++)
		{
			if (pnTest[c] != 0)
				continue;//0-not test cluster
			pnTest[c] = 1;//kept cluster
			for (c1 = c+1; c1 < nClusterNum; c1++)
			{
				if (pnTest[c1] != 0)
					continue;//0-not test cluster
				if (ppdClusterDis[c][c1] <= dDisTh)
				{
					for (n = 0; n < nNum; n++)
					{
						if (pnCluster[n] == c1)
							pnCluster[n] = c;
					}
					pnTest[c1] = 2;
					nClusterNum_bak--;
				}
			}
		}
		bool bCombineCluster = true;
		if (nClusterNum == nClusterNum_bak)
			bCombineCluster = false;
		if (!bCombineCluster)
			break;
		int no = 0;
		memset(pnSampleNum, -1, sizeof(int)*nClusterNum);
		for (c = 0; c < nClusterNum; c++)
		{
			if (pnTest[c] != 2)
			{
				pnSampleNum[c] = no;//use pnSampleNum index new cluster no
				no++;
			}
		}
		assert(no == nClusterNum_bak);
		for (n = 0; n < nNum; n++)
		{
			c = pnCluster[n];
			pnCluster[n] = pnSampleNum[c];
			assert(pnCluster[n]>=0);
		}

		nClusterNum = nClusterNum_bak;
		while (true)
		{//with fixed cluster num update the samples until converge
			// update cluster center first
			skUpdateClusterCenter(pdData, nDim, nNum, pdCC, nClusterNum, pnCluster);
			if (skAssignSample2Cluster(pdData, nDim, nNum, pdWeight, pdCC, nClusterNum, pnCluster) <= 0)
				break;
		}

	}
	for (c = 0; c < nMaxClusterNum; c++)
	{
		delete[] ppdClusterDis[c];
	}
	delete[] ppdClusterDis;
	delete[] pnSampleNum;
	delete[] pnTest;
	return nClusterNum;
}
///////////////////
// skRobustClusterTCA()
//	Author: Cheng-en Guo, 01-26-2002
//	Modified: CG, 11-11-2002
//	input:	pdData - input data, nNum*nDim
//			nDim - sample dimension
//			nNum - sample num
//			pdWeight - weight for each dimension when calculate skEuclidianDistance
//			nMaxClusterNum - supposed max cluster num
//			nMinClusterSize - minimum the sample number in one cluster
//			dMinClusterDisRate - the DisRate=ClusterDis/MaxClusterDis, remove those cluster with lower than a certain threshold
//								If set to 0.0, no consideration
//			nClusterDim - cluster dimension (only part of the original data is used for clustering)
//	output:	pdCC - cluster center, should be nClusterNum*nClusterDim dimension
//			nClusterNum - true cluster num
//			pnCluster - final index of cluster for each sample, should be nNum*1 dimension
//			pnOffset - final offset of the data being used
//	return: cluster num if successed, otherwise 0
int skRobustClusterTCA(double *pdData, int nDataDim, int nNum, double *pdWeight, 
					int nMaxClusterNum, int nMinClusterSize, double dMinClusterDisRate, int nClusterDim, 
					double *pdCC, int *pnCluster, int *pnOffset)
{
	int nClusterNum = nMaxClusterNum;

	if (nDataDim <= 0 || nNum <= 0)
	{ 
		return 0;
	}


	int nDeltaDim = nDataDim - nClusterDim;

	if (nClusterNum <= 0)
		nClusterNum = nMaxClusterNum;

	if (nNum < nMinClusterSize)
		nClusterNum = 1;
	
	if (nNum < nClusterNum)
		nClusterNum = 1;

	int i;//index of sample
	int d;//index of dimension
	int n;//index of data
	int c;//index of cluster
	int o;//index of offset
/*
	// if only one cluster desired
	if (nClusterNum ==  1)
	{
		for (d = 0; d < nClusterDim; d++)
			pdCC[d] = 0;
		n  = 0;
		for (i = 0; i < nNum; i++)
			for (d = 0; d < nClusterDim; d++)
				pdCC[d] += pdData[n++];
		for (d = 0; d < nDim; d++)
			pdCC[d] /= nNum;
		for (i = 0; i < nNum; i++)
			pnCluster[i] = 0;//only one cluster
		return 1;
	}
*/
	// initilize the the cluster center
	// This is a greedy method. Different init may lead to different results
	for (c = 0; c < nClusterNum; c++)
	{
		i = (int)(rand()/(RAND_MAX+0.1) * nNum);
		for (d = 0; d < nClusterDim; d++)
			pdCC[c*nClusterDim + d] = pdData[i*nClusterDim + d + nDeltaDim/2];
	}

	int	   *pnSampleNum = new int[nMaxClusterNum];//the sample num for each cluster
	memset(pnCluster, -1, sizeof(int)*nNum);
	memset(pnOffset, nDeltaDim/2, sizeof(int)*nNum);
	// assign each sample to a cluster
	int *pnTest = new int[nClusterNum];
	//combine short skEuclidianDistance clusters
	double **ppdClusterDis = new double*[nMaxClusterNum];
	for (c = 0; c < nClusterNum; c++)
	{
		ppdClusterDis[c] = new double[nClusterNum];
		memset(ppdClusterDis[c], 0, sizeof(double)*nClusterNum);
	}
	while (true)
	{//with fixed cluster num update the samples until converge
		if (skAssignSample2ClusterTCA(pdData, nDataDim, nNum, pdWeight, pdCC, nClusterDim, nClusterNum, pnCluster, pnOffset) <= 0)
			break;
		skUpdateClusterCenterTCA(pdData, nDataDim, nNum, pdCC, nClusterDim, nClusterNum, pnCluster, pnOffset);
	}
	while (true)
	{
		// recompute the sample num in each cluster
		memset(pnSampleNum, 0, sizeof(int)*nMaxClusterNum);
		for (n = 0; n < nNum; n++)
		{
			pnSampleNum[pnCluster[n]]++;
		}
		// remove small size cluster
		bool bSmallCluster = false;
		for (c = 0; c < nClusterNum; c++)
		{
			if (pnSampleNum[c] < nMinClusterSize)
			{
				memcpy(pdCC+nClusterDim*c, pdCC+nClusterDim*(nClusterNum-1), sizeof(double)*nClusterDim);//copy cluster center
				for (n = 0; n < nNum; n++)
				{
					if (pnCluster[n] == nClusterNum - 1)
						pnCluster[n] = c;
					if (pnCluster[n] == c)
						pnCluster[n] = -1;
				}
				c--;
				nClusterNum--;
				bSmallCluster = true;
				break;
			}
		}
		if (bSmallCluster)
		{
			while (true)
			{//with fixed cluster num update the samples until converge
				if (skAssignSample2ClusterTCA(pdData, nDataDim, nNum, pdWeight, pdCC, nClusterDim, nClusterNum, pnCluster, pnOffset) <= 0)
					break;
				skUpdateClusterCenterTCA(pdData, nDataDim, nNum, pdCC, nClusterDim, nClusterNum, pnCluster, pnOffset);
			}
		}
//		if ( dMinClusterDisRate >= 0.99)
//			break;
		// compute all the distances and find the min and max
		double dDis, dMaxDis = -10000000;
		double dMinDis = 1000000000000;
		int c1;
		for (c = 0; c < nClusterNum-1; c++)
		{
			for (c1 = c+1; c1 < nClusterNum; c1++)
			{
				dDis = skEuclidianDistance(pdCC+c*nClusterDim, pdCC+c1*nClusterDim, nClusterDim, pdWeight);
				ppdClusterDis[c][c1] = dDis;
				if (dMaxDis < dDis)
					dMaxDis = dDis;
				if (dMinDis > dDis)
					dMinDis = dDis;
			}
		}

		double dDisTh = (dMaxDis)*dMinClusterDisRate;
		memset(pnTest, 0, sizeof(int)*nClusterNum);
		int nClusterNum_bak = nClusterNum;
		for (c = 0; c < nClusterNum-1; c++)
		{
			if (pnTest[c] != 0)
				continue;//0-not test cluster
			pnTest[c] = 1;//kept cluster
			for (c1 = c+1; c1 < nClusterNum; c1++)
			{
				if (pnTest[c1] != 0)
					continue;//0-not test cluster
				if (ppdClusterDis[c][c1] <= dDisTh)
				{
					for (n = 0; n < nNum; n++)
					{
						if (pnCluster[n] == c1)
							pnCluster[n] = c;
					}
					pnTest[c1] = 2;
					nClusterNum_bak--;
				}
			}
		}
		bool bCombineCluster = true;
		if (nClusterNum == nClusterNum_bak)
			bCombineCluster = false;
		if (!bCombineCluster)
			break;
		int no = 0;
		memset(pnSampleNum, -1, sizeof(int)*nClusterNum);
		for (c = 0; c < nClusterNum; c++)
		{
			if (pnTest[c] != 2)
			{
				pnSampleNum[c] = no;//use pnSampleNum index new cluster no
				no++;
			}
		}
		assert(no == nClusterNum_bak);
		for (n = 0; n < nNum; n++)
		{
			c = pnCluster[n];
			pnCluster[n] = pnSampleNum[c];
			assert(pnCluster[n]>=0);
		}

		nClusterNum = nClusterNum_bak;
		while (true)
		{//with fixed cluster num update the samples until converge
			// update cluster center first
			skUpdateClusterCenterTCA(pdData, nDataDim, nNum, pdCC, nClusterDim, nClusterNum, pnCluster, pnOffset);
			if (skAssignSample2ClusterTCA(pdData, nDataDim, nNum, pdWeight, pdCC, nClusterDim, nClusterNum, pnCluster, pnOffset) <= 0)
				break;
		}

	}//while true
	for (c = 0; c < nMaxClusterNum; c++)
	{
		delete[] ppdClusterDis[c];
	}
	delete[] ppdClusterDis;
	delete[] pnSampleNum;
	delete[] pnTest;
	return nClusterNum;
}


double skEuclidianDistance(double *pData1, double *pData2, int nDim, double *pdWeight)
{
	double dDis = 0;
	for (int d = 0; d < nDim; d++)
		dDis += (pData1[d]-pData2[d])*(pData1[d]-pData2[d])*pdWeight[d]*pdWeight[d];
	return sqrt(dDis);
}

int skUpdateClusterCenter(double *pdData, int nDim, int nNum, double *pdCC, int nClusterNum, int *pnCluster)
{
	if (nClusterNum <= 0)
		return 1;
	int c, d, i, n;

	// set cluster center to zero
//	for (c = 0; c < nClusterNum; c++)
//		for (d = 0; d < nDim; d++)
//			pdCC[d] = 0;
	memset(pdCC, 0, sizeof(double)*nDim*nClusterNum);

	int pnNum[1000];
	memset(pnNum, 0, sizeof(int)*1000);

	// add up center value
	n = 0;
	for (i = 0; i < nNum; i++)
	{ 
		c = pnCluster[i];
		pnNum[c]++;
		assert(c >= 0 && c < nClusterNum);
		for (d = 0; d < nDim; d++)
			pdCC[c*nDim+d] += pdData[n++];
	}
	// divided by sum
	for (c = 0; c < nClusterNum; c++)
	{
		if (pnNum[c] > 0)
			for (d = 0; d < nDim; d++)
				pdCC[c*nDim+d] /= pnNum[c];
	}
	return 0;
}
// return the number of samples whose label are changed
int skAssignSample2Cluster(double *pdData, int nDim, int nNum, double *pdWeight, double *pdCC, int nClusterNum, int *pnCluster)
{
	if (nClusterNum <= 0)
		return 0;
	int c, i;
	
	double dDis;
	double dMinDis = 0;
	int nMin;
	int sum = 0;
	for (i = 0; i < nNum; i++)
	{
		dMinDis = 100000;
		nMin = -1;
		for (c = 0; c < nClusterNum; c++)
		{
			dDis = skEuclidianDistance(pdData+i*nDim, pdCC+c*nDim, nDim, pdWeight);
			if (dMinDis > dDis)
			{
				dMinDis = dDis;
				nMin = c;
			}
		}
		assert(nMin >= 0);
		if (pnCluster[i] != nMin) 
		{
			sum++;
			pnCluster[i] = nMin;
		}
	}	
	return sum;
}

// return the number of samples whose label are changed
int skAssignSample2ClusterTCA(double *pdData, int nDataDim, int nNum, double *pdWeight, 
						   double *pdCC, int nClusterDim, int nClusterNum, int *pnCluster, int *pnOffset)
{
	if (nClusterNum <= 0)
		return 0;
	int nDeltaDim = nDataDim - nClusterDim;
	int c, i, off, d;
	
	double pdDataTmp[1024];
	double dDis;
	double dMinDis = 0;
	int nMin_c, nMin_off;
	int sum = 0;
	for (i = 0; i < nNum; i++)
	{//for each sample
		dMinDis = 100000;
		nMin_c = -1;
		for (c = 0; c < nClusterNum; c++)
		{//for each cluster
			for (off = -nDeltaDim; off <= nDeltaDim-1; off++)
			{
				if (off >= 0)
				{
					for (d = 0; d < nClusterDim; d++)
						pdDataTmp[d] = pdData[i*nDataDim+d+off];
				}
				else
				{
					for (d = 0; d < nClusterDim; d++)
						pdDataTmp[d] = pdData[(i+1)*nDataDim-d+off];
				}
		
				dDis = skEuclidianDistance(pdDataTmp, pdCC+c*nClusterDim, nClusterDim, pdWeight);
				if (dMinDis > dDis)
				{
					dMinDis = dDis;
					nMin_c = c;
					nMin_off = off;
				}
			}//for off
		}//for c
		assert(nMin_c >= 0);
		if (pnCluster[i] != nMin_c || pnOffset[i] != nMin_off) 
		{
			sum++;
			pnCluster[i] = nMin_c;
			pnOffset[i] = nMin_off;
		}
	}//for i
	return sum;
}
int skUpdateClusterCenterTCA(double *pdData, int nDataDim, int nNum,
						   double *pdCC, int nClusterDim, int nClusterNum, int *pnCluster, int *pnOffset)
{
	if (nClusterNum <= 0)
		return 1;
	int c, d, i, n;

	// set cluster center to zero
//	for (c = 0; c < nClusterNum; c++)
//		for (d = 0; d < nDim; d++)
//			pdCC[d] = 0;
	memset(pdCC, 0, sizeof(double)*nClusterDim*nClusterNum);

	int pnNum[1024];
	memset(pnNum, 0, sizeof(int)*1024);

	// add up center value
	n = 0;
	int off;
	for (i = 0; i < nNum; i++)
	{ 
		c = pnCluster[i];
		pnNum[c]++;
		assert(c >= 0 && c < nClusterNum);
		off = pnOffset[i];
		if (off >= 0)
		{
			for (d = 0; d < nClusterDim; d++)
				pdCC[c*nClusterDim+d] += pdData[i*nDataDim+d+off];
		}
		else
		{
			for (d = 0; d < nClusterDim; d++)
				pdCC[c*nClusterDim+d] += pdData[(i+1)*nDataDim-d+off];
		}

	}
	// divided by sum
	for (c = 0; c < nClusterNum; c++)
	{
		if (pnNum[c] > 0)
			for (d = 0; d < nClusterDim; d++)
				pdCC[c*nClusterDim+d] /= pnNum[c];
	}
	return 0;
}
