#include <direct.h>
#include "skCornerDect.h"

#include "skOutput.h"
#include "inlineFunc.h"
#include "skMatrixIplOp.h"
#include "skIOImage.h"


const char g_pchSusanParamName[] = "susan_param.txt";
const char g_pchSusanWorkDir[] = "susan";

SkSusanParam::SkSusanParam()
{
	m_nMode = -1;
	m_fDistance = 2.0;
	m_nBrightnessStart = 5;
	m_nBrightnessEnd = 50;
	m_nBrightnessStep = 5;
}
int SkSusanParam::read(const char *name)
{
	FILE *pFile;
	pFile = fopen(name, "r");
	if (pFile == 0)
		return 1;
	
	fscanf(pFile, "Mode:                %d\n", &m_nMode);// mode 0~smooth, 1~edge, 2~corner, -1 all
	fscanf(pFile, "Distance:            %f\n", &m_fDistance);// 1.0 (3*3) or 2.0 (37)
	fscanf(pFile, "Brightness start:    %d\n", &m_nBrightnessStart);// start brightness
	fscanf(pFile, "Brightness end:      %d\n", &m_nBrightnessEnd);// end brightness
	fscanf(pFile, "Brightness step:     %d\n", &m_nBrightnessStep);// brightness step
	fclose(pFile);
	return 0;
}

int SkSusanParam::save(const char *name)
{
	FILE *pFile;
	pFile = fopen(name, "w");
	if (pFile == 0)
		return 1;
	
	fprintf(pFile, "Mode:                %d\n", m_nMode);// mode 0~smooth, 1~edge, 2~corner, -1 all
	fprintf(pFile, "Distance:            %.2f\n", m_fDistance);// 1.0 (3*3) or 2.0 (37)
	fprintf(pFile, "Brightness start:    %d\n", m_nBrightnessStart);// start brightness
	fprintf(pFile, "Brightness end:      %d\n", m_nBrightnessEnd);// end brightness
	fprintf(pFile, "Brightness step:     %d\n", m_nBrightnessStep);// brightness step
	fclose(pFile);
	return 0;
}

int skSusanEdgeCornerDetection(const char *pchImgName)
{
	DWORD Time0 = ::GetTickCount();

	char pchDir0[512];
	_getcwd(pchDir0, 512);//get the original directory

	char name[512], pchProjName[512], msg[512], work_dir[32];

	// make the work directory
	strcpy(work_dir, g_pchSusanWorkDir);
	
	//get the project name
	strcpy(pchProjName, pchImgName);
	::removeFileTypeInName(pchProjName);

	//load image
	Matrix<int> pImgChn[3];
	int nChannel = ::skLoadImage2YCrCb(pchImgName, pImgChn);
	if (nChannel <= 0)
	{
		return -1;
	}
	sprintf(msg, "Image loaded: %s\n", pchImgName);
	skOutput(msg);
	

	// set the parameters
	// mode: 0~smooth, 1~edge, 2~corner
	SkSusanParam param;
	if (param.read(g_pchSusanParamName) != 0)
		param.save(g_pchSusanParamName);

	//create a directory for this image project, and set as current directory
	_mkdir(pchProjName);
	_chdir(pchProjName);

	_mkdir(work_dir);
	
	//save a copy in the new directory
	skSaveImageFromYCrCb(pchImgName, pImgChn, 1);
	
	// only use the intensity
	Matrix<unsigned char> image, img_tmp;
	::skCopyMatrix(pImgChn[0], image);

	int ht, wd;
	image.GetDimension(ht, wd);
	Matrix<unsigned char> uchar255;
	uchar255.SetDimension(ht, wd);
	uchar255.Set(255);


	int brightness;
	float distance = param.m_fDistance;
	int mode;

	char t_no[32];
	int t;
	int t_st = param.m_nBrightnessStart;
	int t_en = param.m_nBrightnessEnd;
	int t_step = param.m_nBrightnessStep;
	for (t = t_st; t <= t_en; t += t_step ) 
	{
		brightness = t;
		::changeNumTo2Char(t_no, t);

		// edge detection
		mode = 1;
		img_tmp = image;
		::skSusanEdgeCornerDetection(img_tmp, mode, distance, brightness);

		// save the results
		img_tmp = uchar255 - img_tmp;
		sprintf(name, "%s\\%s_susan_edge_t%s.bmp", work_dir, pchProjName, t_no);
		::skSaveMatrixAsBitmap(name, img_tmp, 0, -1);

		// corner detection 
		mode = 2;
		img_tmp = image;	
		
		::skSusanEdgeCornerDetection(img_tmp, mode, distance, brightness);

		// save the results
		img_tmp = uchar255 - img_tmp;//{254: valid, 255: invalid}
		sprintf(name, "%s\\%s_susan_corner_t%s.bmp", work_dir, pchProjName, t_no);
		::skSaveMatrixAsBitmap(name, img_tmp, 0, -1);

		::skMarkPointsAsDots(img_tmp, 254);
		sprintf(name, "%s\\%s_susan_corner_blob_t%s.bmp", work_dir, pchProjName, t_no);
		::skSaveMatrixAsBitmap(name, img_tmp, 0, -1);
	}

	//change back to original dir
	_chdir(pchDir0);

	DWORD Time1 = ::GetTickCount();
	DWORD dTime = (Time1 - Time0);
	int nMin = (int)(dTime/60000.0);
	double dSec = (dTime - nMin * 60000)/1000.0;
	sprintf(msg, "\n Total corner detection time for image %s:   %d min %.3fs \n ", pchImgName, nMin, dSec);
	::skOutput(msg);
	return -1;
}

///////////////////
// raster: input image and output image/edge/corner, (0,1) 1 ~ interested points
// mode: 0~smooth, 1~edge, 2~corner
//
// SPATIAL CONTROL: d
//
//  In SUSAN smoothing:
//		d controls the size of the Gaussian mask; its
//		default is 4.0. Increasing d gives more smoothing. 
//		In smoothing, the flat 3 by 3 mask can be used instead of
//		a larger Gaussian mask; this gives low smoothing and fast operation.
//  In edge finding:
//		a fixed flat mask is used, either 37 pixels arranged in a "circle"
//		(default), or a 3 by 3 mask which gives finer detail. 
//  In corner finding:
//		only the larger 37 pixel mask is used; d is not variable.
//
//  BRIGHTNESS CONTROL: t
//
//  In all three algorithms, t can be varied (default=20); this is the
//  main threshold to be varied. It determines the maximum difference in
//  greylevels between two pixels which allows them to be considered
//  part of the same "region" in the image. Thus it can be reduced to
//  give more edges or corners, i.e. to be more sensitive, and vice
//  versa. In smoothing, reducing t gives less smoothing, and vice
//  versa. Set t=10 for the test image available from the SUSAN web
//  page.
void skSusanEdgeCornerDetection(Matrix<uchar> &raster, int mode, const float distance, const int brightness)
{
uchar  *in=0, *bp=0, *mid=0;
float  dt=4.0;
int    *r=0,
       argindex=3,
       bt=20,
       principle=0,
       thin_post_proc=1,
       three_by_three=0,
       drawing_mode=1,
       susan_quick=0,
       max_no_corners=1850,
       max_no_edges=2650,i,
       x_size, y_size;
CORNER_LIST corner_list;

	// input image
	x_size = raster.cols();
	y_size = raster.rows();
	in = new unsigned char[x_size*y_size];
	for (i=0; i<raster.size(); i++)
		in[i] = raster.Data(i);

// {{{ look at options //


	dt = distance;
	if (dt<0)
		three_by_three = 1;
	bt = brightness;
	
	three_by_three = 1;

  if ( (principle==1) && (mode==0) )
    mode=1;

// }}} //
// {{{ main processing//

  switch (mode)
  {
    case 0:
     // {{{ smoothing 

      setup_brightness_lut(&bp,bt,2);
      susan_smoothing(three_by_three,in,dt,x_size,y_size,bp);
      break;

// }}} 
    case 1:
      // {{{ edges 

      r   = new int[x_size * y_size];//(int *) malloc(x_size * y_size * sizeof(int));
      setup_brightness_lut(&bp,bt,6);

      if (principle)
      {
        if (three_by_three)
          susan_principle_small(in,r,bp,max_no_edges,x_size,y_size);
        else
          susan_principle(in,r,bp,max_no_edges,x_size,y_size);
        int_to_uchar((uchar*)r,(int*)in,x_size*y_size);
      }
      else
      {
        mid = new uchar[x_size*y_size];//(uchar *)malloc(x_size*y_size);
        memset (mid,100,x_size * y_size); // note not set to zero 

        if (three_by_three)
          susan_edges_small(in,r,mid,bp,max_no_edges,x_size,y_size);
        else
          susan_edges(in,r,mid,bp,max_no_edges,x_size,y_size);
        if(thin_post_proc)
          susan_thin(r,mid,x_size,y_size);
		for (i=0; i<raster.size(); i++)
			in[i] = 0;
        edge_draw(in,mid,x_size,y_size,drawing_mode);
      }

      break;

// }}} 
    case 2:
      // {{{ corners 

      r   = new int[x_size * y_size];//(int *) malloc(x_size * y_size * sizeof(int));
      setup_brightness_lut(&bp,bt,6);

      if (principle)
      {
        susan_principle(in,r,bp,max_no_corners,x_size,y_size);
        int_to_uchar((uchar*)r,(int*)in,x_size*y_size);
      }
      else
      {
        if(susan_quick)
          susan_corners_quick(in,r,bp,max_no_corners,corner_list,x_size,y_size);
        else
          susan_corners(in,r,bp,max_no_corners,corner_list,x_size,y_size);

		for (i=0; i<raster.size(); i++)
			in[i] = 0;
        corner_draw(in,corner_list,x_size,drawing_mode);
      }

      break;


  }    

	// ouput result
	for (i=0; i<raster.size(); i++)
		raster(i) = in[i];

	if (bp != 0)
		delete [](bp-258);

	if (mid != 0)
		delete []mid;

	if (r != 0)
		delete [] r;
	if (in != NULL)
		delete []in;
	in = NULL;
}
